from z3 import *
from random import randint
import random
import time

communicationCost = Real('communicationCost')
computationCost = Real('computationCost')
communicationCost1 = Real('communicationCost1')
computationCost1 = Real('computationCost1')

nTask = 10
nNodes = 5
nUsers = 1
nConstraints = 1

tasks = [ [ Real("x_%s_%s" % (0, 0)) for c in range(3) ] 
      for r in range(nTask*nUsers) ]

nodes = [ [ Real("x_%s_%s" % (0, 0)) for c in range(6) ] 
      for r in range(nNodes+nUsers-1) ]

#{task, node}
assignement = [ [ Bool("ASING_%s_%s" % (r,c)) for c in range(nNodes+nUsers-1) ] 
      for r in range(nTask*nUsers) ]

#{node, user}
RAMresources = [ [ Real("RAM_%s_%s" % (r,c)) for c in range(nUsers+nUsers-1) ] 
for r in range(nNodes) ]

constraints = [ [ Real("cons_%s" % (0)) for c in range(9) ] 
      for r in range(nConstraints) ]


#data to be transmitted between tasks
relation = [ [ 0 for c in range(nTask) ] 
      for r in range(nTask) ]

relation[1][2] = randint(30000,340000) * 8
relation[2][3] = randint(30000,340000) * 8
relation[3][4] = randint(30000,340000) * 8
relation[4][5] = randint(30000,340000) * 8
relation[5][6] = randint(30000,340000) * 8
relation[3][7] = randint(30000,340000) * 8
relation[7][8] = randint(30000,340000) * 8
relation[7][9] = randint(30000,340000) * 8


batteryLevel = randint(0,100)
aux = 0
for x in range(nUsers):
	for t in range(nTask):
		tasks[t][0] = randint(200,6000000) #CameraCalibration
		tasks[t][1] = randint(10,150)
		tasks[t][2] = x	
	aux = aux + nTask


#[0] max time to be executed
#[1] number of tasks involved in the restriction
#[2...n] index of the involved tasks
constraints[0][0] = 0.034
constraints[0][1] = 7
constraints[0][2] = 0
constraints[0][3] = 1
constraints[0][4] = 2
constraints[0][5] = 3
constraints[0][6] = 4
constraints[0][7] = 5
constraints[0][8] = 6

#{CPU, Bandwidth. Upload Cost, k, RAM, policy}
for x in range(nUsers):
	for n in range(nNodes):
		nodes[x + n][0] = randint(30000000,8000000000)
		nodes[x + n][1] = randint(3000000,700000000)
		nodes[x + n][2] = randint(30000000,8000000000)
		nodes[x + n][3] = random.uniform(0.0000000000009,0.00000009)
		nodes[x + n][4] = randint(2000,32000)

#policy according to the battery level of the user's node

nodes[0][5] = 1

if batteryLevel >= 75:
	for n in range (1, nNodes):
		aux = randint(1,2)
		nodes [n][5] = 0.5
		if aux == 1:
			nodes [n][5] = 1
elif batteryLevel < 75 and batteryLevel >= 50:
	for n in range (1, nNodes):
		aux = randint(1,2)
		nodes [n][5] = 0.25
		if aux == 1:
			nodes [n][5] = 0.75
elif batteryLevel < 50 and batteryLevel >= 25:
	for n in range (1, nNodes):
		aux = randint(1,2)
		nodes [n][5] = 0.25
		if aux == 1:
			nodes [n][5] = 0.5
elif batteryLevel < 25:	
	for n in range (1, nNodes):
		nodes [n][5] = 0

timeBefore = time.time()

opt = Optimize()

opt.add([Sum([If(assignement [t][m] ,1,0) for m in range(nNodes)]) == 1 for t in range(nTask*nUsers)])
opt.add([Sum([If(RAMresources [m][u] >= 0,1,0) for u in range(nUsers)]) == nUsers for m in range(nNodes)])
opt.add([Sum([RAMresources [m][u] for u in range(nUsers)]) <= nodes[m][4] for m in range(nNodes)])

#tasks t1,t3 and t7 are assigned to the user's node
opt.add(Sum([If(assignement [0 + nTask*u][u],1,0)* If(assignement [2 + nTask*u][u],1,0) * If(assignement [6 + nTask*u][u],1,0)  for u in range(nUsers)]) == nUsers)

#RAM assignement for each node and user
for u in range(0, nUsers):
	opt.add([Sum([assignement[t+nTask*u][m]*tasks[t+nTask*u][1] for t in range(nTask)]) <= RAMresources[m][u] for m in range(nNodes)])

#time restrictions
for c in range(0, nConstraints):
	opt.add([(Sum([simplify(assignement[constraints[c][i]+(u*nTask)][m]*If(assignement[constraints[c][j]+(u*nTask)][m],0,1)*relation[constraints[c][i]][constraints[c][j]]*1000000/nodes[m][1])for j in range(2, 2 + constraints[c][1])]) + (assignement[constraints[c][i]+(u*nTask)][m]*tasks[constraints[c][i]+(u*nTask)][0]/nodes[m][0]))*1000000 <= constraints[c][0]*1000000 for i in range(2, 2 + constraints[c][1])  for m in range(nNodes) for u in range(nUsers)])

opt.add(communicationCost == Sum([Sum([Sum([Sum([simplify(assignement[i + u*nTask][m]*If(assignement[j+ u*nTask][m],0,1)*(nodes[m][2]*(relation[i][j]/nodes[m][1])*nodes[m][5])) for j in range(nTask)]) for i in range(nTask)]) for u in range(nUsers)]) for m in range(nNodes)]))
opt.add(computationCost == Sum([Sum([simplify(assignement[i][m] * nodes[m][3] * tasks[i][0] * (nodes[m][0]**2) *nodes[m][5]) for i in range(nTask*nUsers)]) for m in range(nNodes)]))
opt.minimize(communicationCost + computationCost)

opt.add(communicationCost1 == Sum([Sum([Sum([simplify(assignement[i + u*nTask][0]*If(assignement[j+ u*nTask][0],0,1)*(nodes[0][2]*(relation[i][j]/nodes[0][1])*nodes[0][5])) for j in range(nTask)]) for i in range(nTask)]) for u in range(nUsers)]))
opt.add(computationCost1 == Sum(Sum([simplify(assignement[i][0] * nodes[0][3] * tasks[i][0] * (nodes[0][0]**2) *nodes[0][5]) for i in range(nTask*nUsers)])))

	
print(opt.check())
m = opt.model()
timeAfter = time.time()
print(m)
print('\n')

print('****RESULTS****')
print('Number of nodes: ' + str(nNodes))
print('Number of tasks: ' + str(nTask))
print('Battery level (user node): ' + str(batteryLevel) + '%')
print('Energy consumption in the DI: ' + str(m[computationCost].numerator_as_long()/m[computationCost].denominator_as_long() + m[communicationCost].numerator_as_long()/m[communicationCost].denominator_as_long()) + ' (J)')
print('Energy consumption in the user node: ' + str(m[computationCost1].numerator_as_long()/m[computationCost1].denominator_as_long() + m[communicationCost1].numerator_as_long()/m[communicationCost1].denominator_as_long()) + ' (J)')
totalTime = timeAfter - timeBefore
print('Execution time: ' + str(totalTime) + ' (s)')